#!/usr/bin/env python
#
# StuDNA tools, version 2009-Jul
# (C) Martin Saturka www.bioplexity.org
# Available under the MIT/X license:
#
# Extracts requested sequences from the sqlite database
# prepared by the 'studna-red' and 'studna-green' scripts.
#

import sys, os, time
import getopt
import string

withsqlite3 = 1
try:
  import sqlite3 as S
except:
  withsqlite3 = 0
if 0 == withsqlite3:
  try:
    from pysqlite2 import dbapi2 as S
  except:
    print("sqlite3 support has to be available")
    sys.exit(1)

class StudnaYellow(object):
  def __init__(self):
    self.configfn = ""
    self.totalcount = 0
    self.linelength = 70

    self.sqlitename = ""
    self.fastaname = ""

    self.error = 0

  def work(self):
    self.totalcount = 0
    self.linelength = 70
    self.sqlitename = ""
    self.fastaname = ""

    if not 0 == self.error:
      return False
    if 0 == len(configfn):
      self.error = 1
      return False

    cnfname = ""
    cnfsave = ""
    cnffrom = 0
    cnfupto = 0
    cnfline = 0
    cnftaxs = []
    cnflens = {}
    cnfgids = {}
    cnfsyms = {}

    trtable = string.maketrans(", \t", ";;;")
    isok = True
    cnfstate = 0

    cnfh = []
    try:
      cnfh = open(configfn)
    except:
      self.error = 2
      return False

    try:
      for line in cnfh:
        if line.startswith("#"):
          continue
        line = line.strip()
        if 0 == len(line):
          continue
        if line.startswith("["):
          ldata = line.strip("[] \t")
          if ldata.isdigit():
            if not "0" == ldata:
              cnfstate = ldata
              if not ldata in cnftaxs:
                cnftaxs.append(ldata)
                cnflens[ldata] = []
                cnfgids[ldata] = []
                cnfsyms[ldata] = []
              continue
          isok = False
          self.error = 101
          break
        ldata = line.split(":")
        if not 2 == len(ldata):
          isok = False
          self.error = 102
          break
        ldata0 = ldata[0].lower().strip()
        ldata1 = ldata[1].strip()
        if 0 == cnfstate:
          if "load" == ldata0:
            cnfname = ldata1
          elif "save" == ldata0:
            cnfsave = ldata1
          elif "from" == ldata0:
            if ldata1.isdigit():
              cnffrom = int(ldata1)
            else:
              if ldata1.startswith("-") and ldata1[1:].isdigit():
                cnffrom = int(ldata1)
          elif "upto" == ldata0:
            if ldata1.isdigit():
              cnfupto = int(ldata1)
            else:
              if ldata1.startswith("-") and ldata1[1:].isdigit():
                cnfupto = int(ldata1)
          elif "line" == ldata0:
            if ldata1.isdigit():
              cnfline = int(ldata1)
          else:
            isok = False
            self.error = 103
            break
        else:
          ldata1 = ldata1.translate(trtable)
          curlist = ldata1.split(";")
          if "gids" == ldata0:
            for item in curlist:
              value = item.strip()
              if (0 < len(value)) and value.isdigit():
                cnfgids[cnfstate].append(value)
          elif "syms" == ldata0:
            for item in curlist:
              value = item.strip()
              if (0 < len(value)):
                cnfsyms[cnfstate].append(value.upper())
          else:
            isok = False
            self.error = 104
            break

    except:
      self.error = 3
      cnfh.close()
      return False

    try:
      cnfh.close()
    except:
      self.error = 4
      return False

    if not isok:
      return False

    if "" == cnfname:
      self.error = 201
      return False

    if "" == cnfsave:
      self.error = 202
      return False

    if 0 == cnffrom:
      self.error = 203
      return False

    if 0 == cnfupto:
      self.error = 204
      return False

    if cnffrom > cnfupto:
      self.error = 205
      return False

    cnfgidcount = 0
    cnfsymcount = 0
    for onetax in cnftaxs:
      gidlen = len(cnfgids[onetax])
      symlen = len(cnfsyms[onetax])

      cnflens[onetax] = [gidlen, symlen]
      cnfgidcount += gidlen
      cnfsymcount += symlen

    if (0 == (cnfgidcount + cnfsymcount)):
      self.error = 205
      return False

    self.sqlitename = cnfname
    self.fastaname = cnfsave

    if 0 < cnfline:
      self.linelength = cnfline

    ffah = []
    try:
      ffah = open(self.fastaname, "w")
    except:
      self.error = 401
      return False

    fheader = "#StuDNA-yellow selected gene upstream sequences\n"
    try:
      ffah.write(fheader)
    except:
      ffah.close()
      self.error = 402
      return False

    try:
      ffah.close()
    except:
      self.error = 403
      return False


    dbcon = []
    dbcur = []
    try:
      dbcon = S.connect(self.sqlitename)
      dbcur = dbcon.cursor()
    except:
      self.error = 301
      return False

    [getacol, setacol, uplecol, dolecol] = [0, 1, 2, 3]
    for onetax in cnftaxs:
      gidlen = cnflens[onetax][0]
      symlen = cnflens[onetax][1]
      if 0 == (gidlen + symlen):
        continue

      qstring = "SELECT genetable, sequtable, uplength, downlength FROM globinfo "
      qstring += "WHERE (genestate = 1) AND (sequstate = 1) AND (taxon = ?) LIMIT 1"
      qparams = (onetax,)

      taxpars = {}
      isavail = False
      try:
        dbcur.execute(qstring, qparams)
        for onerow in dbcur:
          taxpars["gtabname"] = onerow[getacol]
          taxpars["stabname"] = onerow[setacol]
          taxpars["uplength"] = int(onerow[uplecol])
          taxpars["dnlength"] = int(onerow[dolecol])
          isavail = True

      except:
        dbcur.close()
        dbcon.close()
        self.error = 303
        return False

      if not isavail:
        continue

      gtabname = taxpars["gtabname"]
      stabname = taxpars["stabname"]
      uplength = taxpars["uplength"]
      dnlength = taxpars["dnlength"]

      limupl = -1 * uplength

      curfrom = cnffrom
      if 0 > cnffrom:
        if limupl > cnffrom:
          curfrom = limupl
      else:
        if dnlength < cnffrom:
          isavail = False

      curupto = cnfupto
      if 0 > cnfupto:
        if limupl > cnfupto:
          curupto = limupl
      else:
        if dnlength < cnfupto:
          isavail = False

      if curfrom > curupto:
        isavail = False
      if not isavail:
        continue

      sqoffset = 0
      if (0 > curfrom):
        sqoffset = uplength + curfrom
      else:
        sqoffset = uplength + (curfrom - 1)
      sqlength = curupto - curfrom
      if (0 < curfrom) or (0 > curupto):
        sqlength += 1

      usegenes = {}
      usegids = []
      curgids = cnfgids[onetax]
      cursyms = cnfsyms[onetax]

      [genecol, namecol, symbcol] = [0, 1, 2]
      if 0 < gidlen:
        qstring = "SELECT gene, name, symbol FROM " + gtabname + " WHERE gene IN ("

        qmarks = []
        for item in range(gidlen):
          qmarks.append("?")
        qstring += ",".join(qmarks)
        qstring += ")"

        qparams = tuple(curgids)
        try:
          dbcur.execute(qstring, qparams)
          for onerow in dbcur:
            onegene = int(onerow[genecol])
            onename = onerow[namecol]
            onesymb = onerow[symbcol]
            if not onegene in usegids:
              usegenes[onegene] = [onename, onesymb]
              usegids.append(onegene)
              while onename in cursyms:
                cursyms.remove(onename)

        except:
          dbcur.close()
          dbcon.close()
          self.error = 304
          return False

      for usecolname in ["name", "symbol"]:
        symlen = len(cursyms)
        if 0 == symlen:
          continue

        qstring = "SELECT gene, name, symbol FROM " + gtabname + " WHERE "
        qstring += usecolname + " IN ("
        qmarks = []
        for item in range(symlen):
          qmarks.append("?")
        qstring += ",".join(qmarks)
        qstring += ")"

        qparams = tuple(cursyms)
        try:
          dbcur.execute(qstring, qparams)
          for onerow in dbcur:
            onegene = int(onerow[genecol])
            onename = onerow[namecol]
            onesymb = onerow[symbcol]
            if not onegene in usegids:
              usegenes[onegene] = [onename, onesymb]
              usegids.append(onegene)
              while onename in cursyms:
                cursyms.remove(onename)

        except:
          dbcur.close()
          dbcon.close()
          self.error = 305
          return False

      usegids.sort()
      for onegid in usegids:
        geneinfo = usegenes[onegid]
        finname = str(geneinfo[0])
        finsymb = str(geneinfo[1])

        qstring = "SELECT gene, upseq, downseq FROM " + stabname + " WHERE "
        qstring += "gene = ? LIMIT 1"
        qparams = (onegid,)

        finsequ = ""
        finavail = False
        [geneseqcol, upseqcol, downseqcol] = [0, 1, 2]
        try:
          dbcur.execute(qstring, qparams)
          for onerow in dbcur:
            onegene = int(onerow[geneseqcol])
            oneupsq = onerow[upseqcol]
            onednsq = onerow[downseqcol]
            onesequ = oneupsq + onednsq
            finsequ = onesequ[sqoffset:(sqoffset + sqlength)]
            finavail = True

        except:
          dbcur.close()
          dbcon.close()
          self.error = 306
          return False

        if not finavail:
          continue

        self.totalcount += 1

        ffah = []
        try:
          ffah = open(self.fastaname, "a")
        except:
          self.error = 501
          dbcur.close()
          dbcon.close()
          return False

        if str(onegid) == finname:
          finname = "-"

        try:
          gheader = ">taxid|" + str(onetax) + "|geneid|" + str(onegid)
          gheader += "|auth|" + finname + "|symbol|" + finsymb
          gheader += "|start|" + str(curfrom) + "|stop|" + str(curupto)
          gheader += "| studna urs" + "\n"
          ffah.write(gheader)
        except:
          self.error = 502
          dbcur.close()
          dbcon.close()
          ffah.close()
          return False

        curseqlength = len(finsequ)
        curlinecount = curseqlength // self.linelength
        if 0 < (curseqlength % self.linelength):
          curlinecount += 1
        for lind in range(curlinecount):
          auxstart = lind * self.linelength
          finrange = finsequ[auxstart:(auxstart + self.linelength)]
          finrange += "\n"

          try:
            ffah.write(finrange)
          except:
            self.error = 503
            dbcur.close()
            dbcon.close()
            ffah.close()
            return False

        try:
          ffah.close()
        except:
          self.error = 504
          dbcur.close()
          dbcon.close()
          return False

    try:
      dbcur.close()
      dbcon.close()
    except:
      self.error = 302
      return False

    return True


#params: -c configuration_file
if ('__main__' == __name__):
  args = sys.argv[1:]
  argl = len(args)
  runstring = "Should be run: studna-yellow.py -c configuration_file\n"

  configfn = ""

  optlist = []
  try:
    optlist, args = getopt.getopt(args, 'c:')
    for opt, arg in optlist:
      if "-c" == opt:
        configfn = arg

  except:
    sys.stderr.write(runstring)
    sys.exit(1)

  isok = True
  if 0 == len(configfn):
    isok = False

  if not isok:
    sys.stderr.write(runstring)
    sys.exit(1)


  theYellow = StudnaYellow()
  theYellow.configfn = configfn

  rv = theYellow.work()
  if not rv:
    resmsg = "work failed, error: " + str(theYellow.error)
    print(resmsg)
    sys.exit(1)

  tcount = str(theYellow.totalcount)
  resmsg = "work done, " + tcount + " genes"
  print(resmsg)


